/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Netscape security libraries.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1994-2000
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Dan Kegel <dank@kegel.com>, Google
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "prerror.h"

#include "memio.h"

static PRStatus PR_CALLBACK memio_close(PRFileDesc* fd)
{
    struct memio_secret *secret = (struct memio_secret *)fd->secret;
    memio_buffer_destroy(&secret->readbuf);
    memio_buffer_destroy(&secret->writebuf);
    free(secret);
    return PR_SUCCESS;
}

static PRStatus PR_CALLBACK memio_shutdown(PRFileDesc* fd, PRIntn how)
{
    /* TODO: pass shutdown status to app somehow */
    return PR_SUCCESS;
}

static int PR_CALLBACK memio_recv(PRFileDesc* fd, void* buf, PRInt32 len, 
                                  PRIntn flags, PRIntervalTime timeout)
{
    struct memio_secret *secret;
    struct memio_buffer *b;
    int rv;

    if (flags) {
        PR_SetError(PR_NOT_IMPLEMENTED_ERROR, 0);
        return -1;
    }

    secret = (struct memio_secret*)(fd->secret);
    b = &secret->readbuf;
    rv = memio_buffer_get(b, buf, len);
    if (rv == 0 && ! secret->eof) {
        PR_SetError(PR_WOULD_BLOCK_ERROR, 0);
        return -1;
    }
    return rv;
}

static int PR_CALLBACK memio_read(PRFileDesc* fd, void* buf, PRInt32 len)
{
    /* pull bytes from buffer */
    return memio_recv(fd, buf, len, 0, PR_INTERVAL_NO_TIMEOUT);
}

static int PR_CALLBACK memio_send(PRFileDesc* fd, const void* buf, PRInt32 len,
                                  PRIntn flags, PRIntervalTime timeout)
{
    struct memio_secret *secret;
    struct memio_buffer *b;
    int rv;

    secret = (struct memio_secret*)(fd->secret);
    b = &secret->writebuf;
    rv = memio_buffer_put(b, buf, len);
    if (rv == 0) {
        PR_SetError(PR_WOULD_BLOCK_ERROR, 0);
        return -1;
    }
    return rv;
}

static int PR_CALLBACK memio_write(PRFileDesc* fd, const void* buf, PRInt32 len)
{
    /* append bytes to buffer */
    return memio_send(fd, buf, len, 0, PR_INTERVAL_NO_TIMEOUT);
}

static PRStatus PR_CALLBACK memio_getpeername(PRFileDesc *fd, PRNetAddr *addr)
{
    /* TODO: fail if memio_SetPeerName has not been called */
    struct memio_secret *secret = (struct memio_secret*)(fd->secret);
    memcpy(addr, &secret->peername, sizeof(PRNetAddr));
    return PR_SUCCESS;
}

static PRStatus memio_GetSocketOption(PRFileDesc *fd, PRSocketOptionData *data)
{
    /*
     * Even in the original version for real tcp sockets,
     * PR_SockOpt_Nonblocking is a special case that does not
     * translate to a getsockopt() call
     */
    if (PR_SockOpt_Nonblocking == data->option)
    {
        data->value.non_blocking = PR_TRUE;
        return PR_SUCCESS;
    }
    PR_SetError(PR_OPERATION_NOT_SUPPORTED_ERROR, 0);
    return PR_FAILURE;
}

/* 
 * Implement just the bare minimum number of methods needed to make ssl happy.
 *
 * Oddly, PR_Recv calls ssl_Recv calls ssl_SocketIsBlocking calls 
 * PR_GetSocketOption, so we have to provide an implementation of 
 * PR_GetSocketOption that just says "I'm nonblocking".
 */

static struct PRIOMethods  memio_layer_methods = {
    PR_DESC_LAYERED,           /* Type of file represented (tos)           */
    memio_close,               /* close file and destroy descriptor        */
    memio_read,                /* read up to specified bytes into buffer   */
    memio_write,               /* write specified bytes from buffer        */
    NULL,                      /* determine number of bytes available      */
    NULL,                      /*          ditto, 64 bit                   */
    NULL,                      /* flush all buffers to permanent store     */
    NULL,                      /* position the file to the desired place   */
    NULL,                      /*           ditto, 64 bit                  */
    NULL,                      /* Get information about an open file       */
    NULL,                      /*           ditto, 64 bit                  */
    NULL,                      /* Write segments as described by iovector  */
    NULL,                      /* Connect to the specified (net) address   */
    NULL,                      /* Accept a connection for a (net) peer     */
    NULL,                      /* Associate a (net) address with the fd    */
    NULL,                      /* Prepare to listen for (net) connections  */
    memio_shutdown,            /* Shutdown a (net) connection              */
    memio_recv,                /* Solicit up the the specified bytes       */
    memio_send,                /* Send all the bytes specified             */
    NULL,                      /* Solicit (net) bytes and report source    */
    NULL,                      /* Send bytes to (net) address specified    */
    NULL,                      /* Test the fd to see if it is ready        */
    NULL,                      /* Accept and read on a new (net) fd        */
    NULL,                      /* Transmit at entire file                  */
    NULL,                      /* Get (net) address associated with fd     */
    memio_getpeername,         /* Get peer's (net) address                 */
    NULL,                      /* reserved for future use */
    NULL,                      /* reserved for future use */
    memio_GetSocketOption,     /* Get current setting of specified option  */
    NULL,                      /* Set value of specified option            */
    NULL,                      /* Send a (partial) file with header/trailer*/
    NULL,                      /* Continue a nonblocking connect */
    NULL,                      /* reserved for future use */
    NULL,                      /* reserved for future use */
    NULL,                      /* reserved for future use */
    NULL,                      /* reserved for future use */
};

PRDescIdentity memio_identity = -1;

PRFileDesc *memio_CreateIOLayer(int bufsize)
{
    PRFileDesc *fd;
    struct memio_secret *secret;

    if (memio_identity == -1)
        memio_identity = PR_GetUniqueIdentity("memio");

    fd = PR_CreateIOLayerStub(memio_identity, &memio_layer_methods);
    secret = malloc(sizeof(struct memio_secret));
    memset(secret, 0, sizeof(secret));

    memio_buffer_new(&secret->readbuf, bufsize);
    memio_buffer_new(&secret->writebuf, bufsize);
    fd->secret = (PRFilePrivate *)secret;
    return fd;
}

void memio_SetPeerName(PRFileDesc *fd, const PRNetAddr *peername)
{
    PRFileDesc *memiofd = PR_GetIdentitiesLayer(fd, memio_identity);
    struct memio_secret *secret = (struct memio_secret *) memiofd->secret;
    memcpy(&secret->peername, peername, sizeof(PRNetAddr));
}

void memio_SetEOF(PRFileDesc *fd, PRBool eof)
{
    PRFileDesc *memiofd = PR_GetIdentitiesLayer(fd, memio_identity);
    struct memio_secret *secret = (struct memio_secret *) memiofd->secret;
    secret->eof = eof;
}
