/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Netscape security libraries.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1994-2000
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Dan Kegel <dank@kegel.com>, Google
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/* TODO: nspr-isze it */

#include "memio_buffer.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#ifndef TRUE
#define FALSE 0
#define TRUE 1
#endif

#define min(a,b) (((a) < (b)) ? (a) : (b))

void memio_buffer_new(struct memio_buffer *mb, int size)
{
    mb->head = 0;
    mb->tail = 0;
    mb->bufsize = size;
    mb->buf = malloc(size);
}

void memio_buffer_destroy(struct memio_buffer *mb)
{
    free(mb->buf);
    mb->buf = NULL;
}

/* How many bytes have been put into the buffer */
int memio_buffer_used(const struct memio_buffer *mb)
{
    int n = mb->tail - mb->head;
    if (n < 0) n += mb->bufsize;
    return n;
}

/* How many bytes can be read out of the buffer without wrapping */
int memio_buffer_used_contig(const struct memio_buffer *mb)
{ 
    return (((mb->tail >= mb->head) ? mb->tail : mb->bufsize) - mb->head);
}

/* How many bytes can still be put into the buffer */
int memio_buffer_unused(const struct memio_buffer *mb)
{
    return mb->bufsize - memio_buffer_used(mb) - 1;
}

/* How many bytes can be written into the buffer without wrapping */
int memio_buffer_unused_contig(const struct memio_buffer *mb)
{ 
    if (mb->head > mb->tail) return mb->head - mb->tail - 1;
    return mb->bufsize - mb->tail - (mb->head == 0);
}

/* Is buffer completely empty? */
int memio_buffer_empty(const struct memio_buffer *mb)
{
    return mb->head == mb->tail;
}

/* Is buffer completely full? */
int memio_buffer_full(const struct memio_buffer *mb)
{
    return memio_buffer_unused(mb) == 0;
}

/* Write n bytes into the buffer.  Returns number of bytes written. */
int memio_buffer_put(struct memio_buffer *mb, const char *buf, int n)
{
    int len;
    int transferred = 0;

    /* Handle part before wrap */
    len = min(n, memio_buffer_unused_contig(mb));
    if (len) {
        memcpy(&mb->buf[mb->tail], buf, len);
        mb->tail += len;
        if (mb->tail >= mb->bufsize)
            mb->tail = 0;
        n -= len;
        buf += len;
        transferred += len;

        /* Handle part after wrap */
	len = min(n, memio_buffer_unused_contig(mb));
	if (len) {
	    memcpy(&mb->buf[mb->tail], buf, len);
	    mb->tail += len;
            if (mb->tail >= mb->bufsize)
                mb->tail = 0;
	    transferred += len;
        }
    }

    return transferred;
}


/* Read n bytes from the buffer.  Returns number of bytes read. */
int memio_buffer_get(struct memio_buffer *mb, char *buf, int n)
{
    int len;
    int transferred = 0;

    /* Handle part before wrap */
    len = min(n, memio_buffer_used_contig(mb));
    if (len) {
        memcpy(buf, &mb->buf[mb->head], len);
        mb->head += len;
        if (mb->head >= mb->bufsize)
            mb->head = 0;
        n -= len;
        buf += len;
        transferred += len;

        /* Handle part after wrap */
	len = min(n, memio_buffer_used_contig(mb));
	if (len) {
	    memcpy(buf, &mb->buf[mb->head], len);
	    mb->head += len;
            if (mb->head >= mb->bufsize)
                mb->head = 0;
	    transferred += len;
        }
    }

    return transferred;
}

/*------------------------------------------------------------*/

/* Example of how to use Unix I/O with memio_buffers
 * These are not quite suitable for inclusion in nss
 * since they are unix specific 
 */

/* These functions return number of bytes transferred,
 * or -1 if unix error, or MEMIO_BUFFER_EMPTY if there
 * the buffer itself wasn't ready for I/O
 */

int memio_buffer_send(struct memio_buffer *mb, int fd)
{
        int nb;
	int rv;

        assert(mb->head >= 0 && mb->head < mb->bufsize);
        nb = memio_buffer_used_contig(mb);
        if (!nb) {
                rv = MEMIO_BUFFER_EMPTY;
	} else {
		int nsent;
		nsent = write(fd, &mb->buf[mb->head], nb);
		rv = nsent;
		if (nsent > 0) {
			mb->head += nsent;
			if (mb->head >= mb->bufsize) {
				mb->head = 0;
				nsent = memio_buffer_send(mb, fd);
				if (nsent > 0)
					rv += nsent;
			}
		}
	}
        assert(mb->head >= 0 && mb->head < mb->bufsize);

	return rv;
}

int memio_buffer_recv(struct memio_buffer *mb, int fd)
{
        int nb;
	int rv;

        assert(mb->tail >= 0 && mb->tail < mb->bufsize);
        nb = memio_buffer_unused_contig(mb);
        if (!nb) {
		rv = MEMIO_BUFFER_EMPTY;
	} else {
        	int nreceived;
		nreceived = read(fd, &mb->buf[mb->tail], nb);
		rv = nreceived;
		if (nreceived > 0) {
			mb->tail += nreceived;
			if (mb->tail >= mb->bufsize) {
				mb->tail = 0;
				nreceived = memio_buffer_recv(mb, fd);
				if (nreceived > 0)
					rv += nreceived;
			}
		}
	}
	assert(mb->tail >= 0 && mb->tail < mb->bufsize);

	return rv;
}

/*------------------------------------------------------------*/

/* Even a trivial unit test is very helpful when doing circular buffers. */
/*#define TRIVIAL_SELF_TEST*/
#ifdef TRIVIAL_SELF_TEST

#define TEST_BUFLEN 7

#define CHECKEQ(a, b) { \
    if ((a) != (b)) { \
	printf("%d != %d, Test failed line %d\n", a, b, __LINE__); \
	exit(1); \
    } \
}

int main()
{
	struct memio_buffer mb;
        char buf[100];
        int i;

        memio_buffer_new(&mb, TEST_BUFLEN);

	CHECKEQ(memio_buffer_empty(&mb), TRUE);
	CHECKEQ(memio_buffer_full(&mb), FALSE);
	CHECKEQ(memio_buffer_unused_contig(&mb), TEST_BUFLEN-1);
	CHECKEQ(memio_buffer_unused(&mb), TEST_BUFLEN-1);
	CHECKEQ(memio_buffer_used_contig(&mb), 0);
	CHECKEQ(memio_buffer_used(&mb), 0);

        CHECKEQ(memio_buffer_put(&mb, "howdy", 5), 5);

	CHECKEQ(memio_buffer_empty(&mb), FALSE);
	CHECKEQ(memio_buffer_full(&mb), FALSE);
	CHECKEQ(memio_buffer_unused_contig(&mb), TEST_BUFLEN-1-5);
	CHECKEQ(memio_buffer_unused(&mb), TEST_BUFLEN-1-5);
	CHECKEQ(memio_buffer_used_contig(&mb), 5);
	CHECKEQ(memio_buffer_used(&mb), 5);

        CHECKEQ(memio_buffer_put(&mb, "!", 1), 1);

	CHECKEQ(memio_buffer_empty(&mb), FALSE);
	CHECKEQ(memio_buffer_full(&mb), TRUE);
	CHECKEQ(memio_buffer_unused_contig(&mb), 0);
	CHECKEQ(memio_buffer_unused(&mb), 0);
	CHECKEQ(memio_buffer_used_contig(&mb), 6);
	CHECKEQ(memio_buffer_used(&mb), 6);

        CHECKEQ(memio_buffer_get(&mb, buf, 6), 6);
	CHECKEQ(memcmp(buf, "howdy!", 6), 0);

	CHECKEQ(memio_buffer_empty(&mb), TRUE);
	CHECKEQ(memio_buffer_full(&mb), FALSE);
	CHECKEQ(memio_buffer_unused(&mb), TEST_BUFLEN-1);
	CHECKEQ(memio_buffer_unused_contig(&mb), 1);
	CHECKEQ(memio_buffer_used_contig(&mb), 0);
	CHECKEQ(memio_buffer_used(&mb), 0);

        CHECKEQ(memio_buffer_put(&mb, "01234", 5), 5);

	CHECKEQ(memio_buffer_empty(&mb), FALSE);
	CHECKEQ(memio_buffer_full(&mb), FALSE);
	CHECKEQ(memio_buffer_used(&mb), 5);
	CHECKEQ(memio_buffer_used_contig(&mb), 1);
	CHECKEQ(memio_buffer_unused_contig(&mb), TEST_BUFLEN-1-5);
	CHECKEQ(memio_buffer_unused(&mb), TEST_BUFLEN-1-5);

        CHECKEQ(memio_buffer_put(&mb, "5", 1), 1);

	CHECKEQ(memio_buffer_empty(&mb), FALSE);
	CHECKEQ(memio_buffer_full(&mb), TRUE);
	CHECKEQ(memio_buffer_unused_contig(&mb), 0);
	CHECKEQ(memio_buffer_unused(&mb), 0);
	CHECKEQ(memio_buffer_used_contig(&mb), 1);
	CHECKEQ(memio_buffer_used(&mb), 6);

	/* TODO: add more cases */

	printf("Test passed\n");
	exit(0);
}

#endif
