/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Netscape security libraries.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1994-2000
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Dr Vipul Gupta <vipul.gupta@sun.com>, Sun Microsystems Laboratories
 *   Douglas Stebila <douglas@stebila.ca>, Sun Microsystems Laboratories
 *   Dan Kegel <dank@kegel.com>, Google
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
**
** Sample client side test program that runs any desired number of
** simultaneous SSL sessions, using native Unix I/O rather than NSPR I/O
**
*/

#include "memio.h"

#include "secutil.h"

#if defined(XP_UNIX)
#include <unistd.h>
#else
#include <ctype.h>	/* for isalpha() */
#endif

#include <event.h>	/* for libevent stuff */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>

#include "nspr.h"
#include "prio.h"
#include "prnetdb.h"
#include "nss.h"
#include "ssl.h"
#include "sslproto.h"
#include "pk11func.h"
#include "plgetopt.h"
#include "plstr.h"

#define PRINTF  if (verbose)  printf
#define FPRINTF if (verbose) fprintf

#define MAX_WAIT_FOR_SERVER 600
#define WAIT_INTERVAL       100
#define MAXSESSIONS 1000

PRIntervalTime maxInterval    = PR_INTERVAL_NO_TIMEOUT;

int ssl2CipherSuites[] = {
    SSL_EN_RC4_128_WITH_MD5,			/* A */
    SSL_EN_RC4_128_EXPORT40_WITH_MD5,		/* B */
    SSL_EN_RC2_128_CBC_WITH_MD5,		/* C */
    SSL_EN_RC2_128_CBC_EXPORT40_WITH_MD5,	/* D */
    SSL_EN_DES_64_CBC_WITH_MD5,			/* E */
    SSL_EN_DES_192_EDE3_CBC_WITH_MD5,		/* F */
    0
};

int ssl3CipherSuites[] = {
    -1, /* SSL_FORTEZZA_DMS_WITH_FORTEZZA_CBC_SHA* a */
    -1, /* SSL_FORTEZZA_DMS_WITH_RC4_128_SHA,	 * b */
    SSL_RSA_WITH_RC4_128_MD5,			/* c */
    SSL_RSA_WITH_3DES_EDE_CBC_SHA,		/* d */
    SSL_RSA_WITH_DES_CBC_SHA,			/* e */
    SSL_RSA_EXPORT_WITH_RC4_40_MD5,		/* f */
    SSL_RSA_EXPORT_WITH_RC2_CBC_40_MD5,		/* g */
    -1, /* SSL_FORTEZZA_DMS_WITH_NULL_SHA,	 * h */
    SSL_RSA_WITH_NULL_MD5,			/* i */
    SSL_RSA_FIPS_WITH_3DES_EDE_CBC_SHA,		/* j */
    SSL_RSA_FIPS_WITH_DES_CBC_SHA,		/* k */
    TLS_RSA_EXPORT1024_WITH_DES_CBC_SHA,	/* l */
    TLS_RSA_EXPORT1024_WITH_RC4_56_SHA,	        /* m */
    SSL_RSA_WITH_RC4_128_SHA,			/* n */
    TLS_DHE_DSS_WITH_RC4_128_SHA,		/* o */
    SSL_DHE_RSA_WITH_3DES_EDE_CBC_SHA,		/* p */
    SSL_DHE_DSS_WITH_3DES_EDE_CBC_SHA,		/* q */
    SSL_DHE_RSA_WITH_DES_CBC_SHA,		/* r */
    SSL_DHE_DSS_WITH_DES_CBC_SHA,		/* s */
    TLS_DHE_DSS_WITH_AES_128_CBC_SHA, 	    	/* t */
    TLS_DHE_RSA_WITH_AES_128_CBC_SHA,       	/* u */
    TLS_RSA_WITH_AES_128_CBC_SHA,     	    	/* v */
    TLS_DHE_DSS_WITH_AES_256_CBC_SHA, 	    	/* w */
    TLS_DHE_RSA_WITH_AES_256_CBC_SHA,       	/* x */
    TLS_RSA_WITH_AES_256_CBC_SHA,     	    	/* y */
    SSL_RSA_WITH_NULL_SHA,			/* z */
    0
};

unsigned long __cmp_umuls;
PRBool verbose;
int renegotiate = 0;

static char *progName;

/* This exists only for the automated test suite. It allows us to
 * pass in a password on the command line. 
 */

char *password = NULL;

char * ownPasswd( PK11SlotInfo *slot, PRBool retry, void *arg)
{
	char *passwd = NULL;
	if ( (!retry) && arg ) {
		passwd = PL_strdup((char *)arg);
	}
	return passwd;
}

void printSecurityInfo(PRFileDesc *fd)
{
    CERTCertificate * cert;
    SSL3Statistics * ssl3stats = SSL_GetStatistics();
    SECStatus result;
    SSLChannelInfo    channel;
    SSLCipherSuiteInfo suite;

    result = SSL_GetChannelInfo(fd, &channel, sizeof channel);
    if (result == SECSuccess && 
        channel.length == sizeof channel && 
	channel.cipherSuite) {
	result = SSL_GetCipherSuiteInfo(channel.cipherSuite, 
					&suite, sizeof suite);
	if (result == SECSuccess) {
	    FPRINTF(stderr, 
	    "%s: SSL version %d.%d using %d-bit %s with %d-bit %s MAC\n",
	       progName, channel.protocolVersion >> 8, channel.protocolVersion & 0xff,
	       suite.effectiveKeyBits, suite.symCipherName, 
	       suite.macBits, suite.macAlgorithmName);
	    FPRINTF(stderr, 
	    "%s: Server Auth: %d-bit %s, Key Exchange: %d-bit %s\n",
	       progName, channel.authKeyBits, suite.authAlgorithmName,
	       channel.keaKeyBits,  suite.keaTypeName);
    	}
    }
    cert = SSL_RevealCert(fd);
    if (cert) {
	char * ip = CERT_NameToAscii(&cert->issuer);
	char * sp = CERT_NameToAscii(&cert->subject);
        if (sp) {
	    fprintf(stderr, "subject DN: %s\n", sp);
	    PORT_Free(sp);
	}
        if (ip) {
	    fprintf(stderr, "issuer  DN: %s\n", ip);
	    PORT_Free(ip);
	}
	CERT_DestroyCertificate(cert);
	cert = NULL;
    }
    fprintf(stderr,
    	"%ld cache hits; %ld cache misses, %ld cache not reusable\n"
	"%ld stateless resumes\n",
    	ssl3stats->hsh_sid_cache_hits, ssl3stats->hsh_sid_cache_misses,
	ssl3stats->hsh_sid_cache_not_ok, ssl3stats->hsh_sid_stateless_resumes);
}

void
handshakeCallback(PRFileDesc *fd, void *client_data)
{
    printSecurityInfo(fd);
    if (renegotiate > 0) {
	renegotiate--;
	SSL_ReHandshake(fd, PR_FALSE);
    }
}

static void Usage(const char *progName)
{
    fprintf(stderr, 
"Usage:  %s -h host [-p port] [-d certdir] [-n nickname] [-23BTfosvxr] \n"
"                   [-c connections] [-C ciphers] [-w passwd] [-q]\n", progName);
    fprintf(stderr, "%-20s Hostname to connect with\n", "-h host");
    fprintf(stderr, "%-20s Port number for SSL server\n", "-p port");
    fprintf(stderr, 
            "%-20s Directory with cert database (default is ~/.netscape)\n",
	    "-d certdir");
    fprintf(stderr, "%-20s Nickname of key and cert for client auth\n", 
                    "-n nickname");
    fprintf(stderr, 
            "%-20s Bypass PKCS11 layer for SSL encryption and MACing.\n", "-B");
    fprintf(stderr, "%-20s Disable SSL v2.\n", "-2");
    fprintf(stderr, "%-20s Disable SSL v3.\n", "-3");
    fprintf(stderr, "%-20s Disable TLS (SSL v3.1).\n", "-T");
    fprintf(stderr, "%-20s Prints only payload data. Skips HTTP header.\n", "-S");
    fprintf(stderr, "%-20s Client speaks first. \n", "-f");
    fprintf(stderr, "%-20s Override bad server cert. Make it OK.\n", "-o");
    fprintf(stderr, "%-20s Disable SSL socket locking.\n", "-s");
    fprintf(stderr, "%-20s Verbose progress reporting.\n", "-v");
    fprintf(stderr, "%-20s Use export policy.\n", "-x");
    fprintf(stderr, "%-20s Ping the server and then exit.\n", "-q");
    fprintf(stderr, "%-20s Renegotiate with session resumption.\n", "-r");
    fprintf(stderr, "%-20s Enable the session ticket extension.\n", "-u");
    fprintf(stderr, "%-20s Letter(s) chosen from the following list\n", 
                    "-C ciphers");
    fprintf(stderr, 
"A    SSL2 RC4 128 WITH MD5\n"
"B    SSL2 RC4 128 EXPORT40 WITH MD5\n"
"C    SSL2 RC2 128 CBC WITH MD5\n"
"D    SSL2 RC2 128 CBC EXPORT40 WITH MD5\n"
"E    SSL2 DES 64 CBC WITH MD5\n"
"F    SSL2 DES 192 EDE3 CBC WITH MD5\n"
"\n"
"c    SSL3 RSA WITH RC4 128 MD5\n"
"d    SSL3 RSA WITH 3DES EDE CBC SHA\n"
"e    SSL3 RSA WITH DES CBC SHA\n"
"f    SSL3 RSA EXPORT WITH RC4 40 MD5\n"
"g    SSL3 RSA EXPORT WITH RC2 CBC 40 MD5\n"
"i    SSL3 RSA WITH NULL MD5\n"
"j    SSL3 RSA FIPS WITH 3DES EDE CBC SHA\n"
"k    SSL3 RSA FIPS WITH DES CBC SHA\n"
"l    SSL3 RSA EXPORT WITH DES CBC SHA\t(new)\n"
"m    SSL3 RSA EXPORT WITH RC4 56 SHA\t(new)\n"
"n    SSL3 RSA WITH RC4 128 SHA\n"
"o    SSL3 DHE DSS WITH RC4 128 SHA\n"
"p    SSL3 DHE RSA WITH 3DES EDE CBC SHA\n"
"q    SSL3 DHE DSS WITH 3DES EDE CBC SHA\n"
"r    SSL3 DHE RSA WITH DES CBC SHA\n"
"s    SSL3 DHE DSS WITH DES CBC SHA\n"
"t    SSL3 DHE DSS WITH AES 128 CBC SHA\n"
"u    SSL3 DHE RSA WITH AES 128 CBC SHA\n"
"v    SSL3 RSA WITH AES 128 CBC SHA\n"
"w    SSL3 DHE DSS WITH AES 256 CBC SHA\n"
"x    SSL3 DHE RSA WITH AES 256 CBC SHA\n"
"y    SSL3 RSA WITH AES 256 CBC SHA\n"
"z    SSL3 RSA WITH NULL SHA\n"
"\n"
":WXYZ  Use cipher with hex code { 0xWX , 0xYZ } in TLS\n"
	);
    exit(1);
}

void
milliPause(PRUint32 milli)
{
    PRIntervalTime ticks = PR_MillisecondsToInterval(milli);
    PR_Sleep(ticks);
}

void
disableAllSSLCiphers(void)
{
    const PRUint16 *cipherSuites = SSL_ImplementedCiphers;
    int             i            = SSL_NumImplementedCiphers;
    SECStatus       rv;

    /* disable all the SSL3 cipher suites */
    while (--i >= 0) {
	PRUint16 suite = cipherSuites[i];
        rv = SSL_CipherPrefSetDefault(suite, PR_FALSE);
	if (rv != SECSuccess) {
	    PRErrorCode err = PR_GetError();
	    fprintf(stderr,
	            "SSL_CipherPrefSet didn't like value 0x%04x (i = %d): %s\n",
	    	   suite, i, SECU_Strerror(err));
	    exit(2);
	}
    }
}

/*
 * Callback is called when incoming certificate is not valid.
 * Returns SECSuccess to accept the cert anyway, SECFailure to reject.
 */
static SECStatus 
ownBadCertHandler(void * arg, PRFileDesc * socket)
{
    PRErrorCode err = PR_GetError();
    /* can log invalid cert here */
    fprintf(stderr, "Bad server certificate: %d, %s\n", err, 
            SECU_Strerror(err));
    return SECSuccess;	/* override, say it's OK. */
}

SECStatus
own_GetClientAuthData(void *                       arg,
                      PRFileDesc *                 socket,
                      struct CERTDistNamesStr *    caNames,
                      struct CERTCertificateStr ** pRetCert,
                      struct SECKEYPrivateKeyStr **pRetKey)
{
    if (verbose > 0) {
	SECStatus rv;
        fprintf(stderr, "Server requested Client Authentication\n");
	if (caNames && caNames->nnames > 0) {
	    PLArenaPool *arena = caNames->arena;
	    if (!arena)
		arena = PORT_NewArena(2048);
	    if (arena) {
		int i;
		for (i = 0; i < caNames->nnames; ++i) {
		    char *nameString;
		    CERTName dn;
		    rv = SEC_QuickDERDecodeItem(arena, 
					    &dn,
					    SEC_ASN1_GET(CERT_NameTemplate), 
					    caNames->names + i);
		    if (rv != SECSuccess)
			continue;
		    nameString = CERT_NameToAscii(&dn);
		    if (!nameString)
			continue;
		    fprintf(stderr, "CA[%d]: %s\n", i + 1, nameString);
		    PORT_Free(nameString);
		}
		if (!caNames->arena) {
		    PORT_FreeArena(arena, PR_FALSE);
		}
	    }
	}
	rv = NSS_GetClientAuthData(arg, socket, caNames, pRetCert, pRetKey);
	if (rv == SECSuccess && *pRetCert) {
	    char *nameString = CERT_NameToAscii(&((*pRetCert)->subject));
	    if (nameString) {
		fprintf(stderr, "sent cert: %s\n", nameString);
		PORT_Free(nameString);
	    }
	} else {
	    fprintf(stderr, "send no cert\n");
	}
	return rv;
    }
    return NSS_GetClientAuthData(arg, socket, caNames, pRetCert, pRetKey);
}

static void
printHostNameAndAddr(const char * host, const PRNetAddr * addr)
{
    PRUint16 port = PR_NetAddrInetPort(addr);
    char addrBuf[80];
    PRStatus st = PR_NetAddrToString(addr, addrBuf, sizeof addrBuf);

    if (st == PR_SUCCESS) {
	port = PR_ntohs(port);
	FPRINTF(stderr, "%s: connecting to %s:%hu (address=%s)\n",
	       progName, host, port, addrBuf);
    }
}

/*
 *  Prints output according to skipProtoHeader flag. If skipProtoHeader
 *  is not set, prints without any changes, otherwise looking
 *  for \n\r\n(empty line sequence: HTTP header separator) and
 *  prints everything after it.
 */
static void
separateReqHeader(const PRFileDesc* outFd, const char* buf, const int nb,
                  PRBool *wrStarted, int *ptrnMatched) {

    /* it is sufficient to look for only "\n\r\n". Hopping that
     * HTTP response format satisfies the standard */
    char *ptrnStr = "\n\r\n";
    char *resPtr;

    if (nb == 0) {
        return;
    }

    if (*ptrnMatched > 0) {
        /* Get here only if previous separateReqHeader call found
         * only a fragment of "\n\r\n" in previous buffer. */
        PORT_Assert(*ptrnMatched < 3);

        /* the size of fragment of "\n\r\n" what we want to find in this
         * buffer is equal to *ptrnMatched */
        if (*ptrnMatched <= nb) {
            /* move the pointer to the beginning of the fragment */
            int strSize = *ptrnMatched;
            char *tmpPtrn = ptrnStr + (3 - strSize);
            if (PL_strncmp(buf, tmpPtrn, strSize) == 0) {
                /* print the rest of the buffer(without the fragment) */
                PR_Write((void*)outFd, buf + strSize, nb - strSize);
                *wrStarted = PR_TRUE;
                return;
            }
        } else {
            /* we are here only when nb == 1 && *ptrnMatched == 2 */
            if (*buf == '\r') {
                *ptrnMatched = 1;
            } else {
                *ptrnMatched = 0;
            }
            return;
        }
    }
    resPtr = PL_strnstr(buf, ptrnStr, nb);
    if (resPtr != NULL) {
        /* if "\n\r\n" was found in the buffer, calculate offset
         * and print the rest of the buffer */
        int newBn = nb - (resPtr - buf + 3); /* 3 is the length of "\n\r\n" */

        PR_Write((void*)outFd, resPtr + 3, newBn);
        *wrStarted = PR_TRUE;
        return;
    } else {
        /* try to find a fragment of "\n\r\n" at the end of the buffer.
         * if found, set *ptrnMatched to the number of chars left to find
         * in the next buffer.*/
        int i;
        for(i = 1 ;i < 3;i++) {
            char *bufPrt;
            int strSize = 3 - i;
            
            if (strSize > nb) {
                continue;
            }
            bufPrt = (char*)(buf + nb - strSize);
            
            if (PL_strncmp(bufPrt, ptrnStr, strSize) == 0) {
                *ptrnMatched = i;
                return;
            }
        }
    }
}

PRNetAddr          addr;
char *             cipherString = NULL;
int                override = 0;
int                disableSSL2 = 0;
int                disableSSL3 = 0;
int                disableTLS  = 0;
int                bypassPKCS11 = 0;
int                disableLocking = 0;
int                useExportPolicy = 0;
int                enableSessionTickets = 0;
PRBool             useCommandLinePassword = PR_FALSE;
PRBool             clientSpeaksFirst = PR_FALSE;
PRBool             wrStarted = PR_FALSE;
PRBool             skipProtoHeader = PR_FALSE;
char *             nickname =  NULL;

CERTCertDBHandle * handle;

#define HEXCHAR_TO_INT(c, i) \
    if (((c) >= '0') && ((c) <= '9')) { \
	i = (c) - '0'; \
    } else if (((c) >= 'a') && ((c) <= 'f')) { \
	i = (c) - 'a' + 10; \
    } else if (((c) >= 'A') && ((c) <= 'F')) { \
	i = (c) - 'A' + 10; \
    } else { \
	Usage(progName); \
    }

PRFileDesc * setup_client(PRFileDesc *s)
{
    SECStatus          rv;

    s = SSL_ImportFD(NULL, s);
    if (s == NULL) {
	SECU_PrintError(progName, "error importing socket");
	return NULL;
    }

    rv = SSL_OptionSet(s, SSL_SECURITY, 1);
    if (rv != SECSuccess) {
        SECU_PrintError(progName, "error enabling socket");
	return NULL;
    }

    rv = SSL_OptionSet(s, SSL_HANDSHAKE_AS_CLIENT, 1);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "error enabling client handshake");
	return NULL;
    }

    /* all the SSL2 and SSL3 cipher suites are enabled by default. */
    if (cipherString) {
    	char *cstringSaved = cipherString;
    	int ndx;

	while (0 != (ndx = *cipherString++)) {
	    int  cipher;

	    if (ndx == ':') {
		int ctmp;

		cipher = 0;
		HEXCHAR_TO_INT(*cipherString, ctmp)
		cipher |= (ctmp << 12);
		cipherString++;
		HEXCHAR_TO_INT(*cipherString, ctmp)
		cipher |= (ctmp << 8);
		cipherString++;
		HEXCHAR_TO_INT(*cipherString, ctmp)
		cipher |= (ctmp << 4);
		cipherString++;
		HEXCHAR_TO_INT(*cipherString, ctmp)
		cipher |= ctmp;
		cipherString++;
	    } else {
		const int *cptr;

		if (! isalpha(ndx))
		    Usage(progName);
		cptr = islower(ndx) ? ssl3CipherSuites : ssl2CipherSuites;
		for (ndx &= 0x1f; (cipher = *cptr++) != 0 && --ndx > 0; ) 
		    /* do nothing */;
	    }
	    if (cipher > 0) {
		SECStatus status;
		status = SSL_CipherPrefSet(s, cipher, SSL_ALLOWED);
		if (status != SECSuccess) 
		    SECU_PrintError(progName, "SSL_CipherPrefSet()");
	    } else {
		Usage(progName);
	    }
	}
	PORT_Free(cstringSaved);
    }

    rv = SSL_OptionSet(s, SSL_ENABLE_SSL2, !disableSSL2);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "error enabling SSLv2 ");
	return NULL;
    }

    rv = SSL_OptionSet(s, SSL_ENABLE_SSL3, !disableSSL3);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "error enabling SSLv3 ");
	return NULL;
    }

    rv = SSL_OptionSet(s, SSL_ENABLE_TLS, !disableTLS);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "error enabling TLS ");
	return NULL;
    }

    /* disable ssl2 and ssl2-compatible client hellos. */
    rv = SSL_OptionSet(s, SSL_V2_COMPATIBLE_HELLO, !disableSSL2);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "error disabling v2 compatibility");
	return NULL;
    }

    /* enable PKCS11 bypass */
    rv = SSL_OptionSet(s, SSL_BYPASS_PKCS11, bypassPKCS11);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "error enabling PKCS11 bypass");
	return NULL;
    }

    /* disable SSL socket locking */
    rv = SSL_OptionSet(s, SSL_NO_LOCKS, disableLocking);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "error disabling SSL socket locking");
	return NULL;
    }

    /* enable Session Ticket extension. */
    rv = SSL_OptionSet(s, SSL_ENABLE_SESSION_TICKETS, enableSessionTickets);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "error enabling Session Ticket extension");
	return NULL;
    }

    if (useCommandLinePassword) {
	SSL_SetPKCS11PinArg(s, password);
    }

    SSL_AuthCertificateHook(s, SSL_AuthCertificate, (void *)handle);
    if (override) {
	SSL_BadCertHook(s, ownBadCertHandler, NULL);
    }
    SSL_GetClientAuthDataHook(s, own_GetClientAuthData, (void *)nickname);
    SSL_HandshakeCallback(s, handshakeCallback, NULL);

    return s;
}

enum demo_state {
    DEMO_START,
    DEMO_CONNECTING,
    DEMO_SENDING,
    DEMO_READING,
    DEMO_DYING,
    DEMO_DEAD
};

struct demo_session {
    int id;
    enum demo_state state;

    /* The SSL state machine */
    PRFileDesc *s;

    /* A buffer for stuff going into the app end of the SSL state machine */
    int bufptr;
    int buflen;
    char buf[1000];

    /* Buffers for the network end of the SSL state machine */
    struct memio_secret *bufs;

    /* Native socket */
    int fd;

    /* libevent struct to help watch for events on fd */
    struct event e;
};

/* PR_GetConnectStatus takes a PRPollDesc*, which is going to make it hard
 * to avoid using PR_Poll.  Work around this by providing a wrapper that 
 * takes a PRFileDesc*.  Alternately just use the native getsockopt.
 */

int PR_GetConnectStatus2(PRFileDesc *fd) 
{
    PRPollDesc dummy;
    memset(&dummy, 0, sizeof(dummy));
    dummy.fd = fd;
    dummy.out_flags |= PR_POLL_WRITE;
    return PR_GetConnectStatus(&dummy);
}

PRFileDesc *       std_out;
int                headerSeparatorPtrnId = 0;

/* Nudge this demo session along its path a step. */
void crank_state_machine(struct demo_session *ds)
{
    char buf[256];
    PRFileDesc *s = ds->s;
    int fd = ds->fd;
    int more;
    int nb;
    int err;
    socklen_t len;

    if (verbose > 2) 
        fprintf(stderr, "%s: %d: crank_state_machine top\n", progName, ds->id);

    do {
        int status;
	int nrecieved;
	more = 0;

        memio_buffer_send(&ds->bufs->writebuf, fd);
        nrecieved = memio_buffer_recv(&ds->bufs->readbuf, fd);
	if (nrecieved == 0)
		memio_SetEOF(s, PR_TRUE);

	switch (ds->state) {
	case DEMO_START:
	    /* Try to connect to the server */
	    FPRINTF(stderr, "%s: %d: connecting\n", progName, ds->id);
	    status = connect(fd, (struct sockaddr *)&addr.inet, 
                             sizeof(struct sockaddr_in));
	    if (status == 0) {
		ds->state = DEMO_SENDING;
		more = 1;
	    } else if (errno == EINPROGRESS) {
		ds->state = DEMO_CONNECTING;
	    } else {
		SECU_PrintError(progName, "unable to connect");
		ds->state = DEMO_DYING;
                more = 1;
	    }
	    break;
	case DEMO_CONNECTING: 
	    err = -1;
	    len = sizeof(err);
	    FPRINTF(stderr, "%s: %d: resume connecting\n", progName, ds->id);
	    if (getsockopt(fd, SOL_SOCKET, SO_ERROR, (char*)&err, &len) < 0) {
		/* yuck, couldn't even retrieve error code... */
		err = errno;
	    }
	    if (err == EINPROGRESS) {
		FPRINTF(stderr, "%s: %d: still connecting\n", progName, ds->id);
		/* Await next callback.  Haven't seen this case yet myself. */
	    } else if (!err) {
		PRNetAddr peername;
		socklen_t len;
		int err;

		/* Success! */
		FPRINTF(stderr, "%s: %d: connect finished\n", progName, ds->id);
		len = sizeof(PRNetAddr);
		err = getpeername( fd, (struct sockaddr *)&peername, &len );
		if (err) {
		    FPRINTF(stderr, "%s: %d: getpeername failed, errno %d\n",
                            progName, ds->id, errno);
		    ds->state = DEMO_DYING;
		    more = 1;
		    break;
		}
		memio_SetPeerName(s, &peername);

		/* Send bytes into SSL state machine */
		ds->bufptr = 0;
		ds->buflen = sprintf(ds->buf, "GET /\n");
		ds->state = DEMO_SENDING;
		more = 1;
	    } else {
		ds->state = DEMO_DYING;
		more = 1;
		FPRINTF(stderr, "%s: %d: connect failed, errno %d\n",
                        progName, ds->id, err);
	    }
            break;
	case DEMO_SENDING:
	    {
		char *bufp = ds->buf + ds->bufptr;
		nb = ds->buflen - ds->bufptr;
		PRInt32 cc = PR_Send(s, bufp, nb, 0, maxInterval);
		if (cc < 0) {
		    PRErrorCode err = PR_GetError();
		    if (err != PR_WOULD_BLOCK_ERROR) {
			SECU_PrintError(progName, "write to SSL socket failed");
			ds->state = DEMO_DYING;
                        more = 1;
			break;
		    }
		    cc = 0;
		} else {
		    FPRINTF(stderr, "%s: %d: Tried to send %d bytes, sent %d\n",
                            progName, ds->id, nb, cc);
		    ds->bufptr += cc;
		    if (ds->bufptr >= ds->buflen)
		        ds->state = DEMO_READING;
		    more = 1;
                }
	    }
            break;
	case DEMO_READING:
	    nb = PR_Recv(s, buf, sizeof buf, 0, maxInterval);
	    if (nb < 0) {
		if (PR_GetError() != PR_WOULD_BLOCK_ERROR) {
		    SECU_PrintError(progName, "read from socket failed");
		    ds->state = DEMO_DYING;
                    more = 1;
		}
	    } else if (nb == 0) {
	        FPRINTF(stderr, "%s: %d: EOF from server\n", progName, ds->id);
		/* EOF from socket... stop polling socket for read */
		ds->state = DEMO_DYING;
                more = 1;
	    } else {
	        FPRINTF(stderr, "%s: %d: Read from server %d bytes:\n[[[\n",
                        progName, ds->id, nb);
		if (skipProtoHeader != PR_TRUE || wrStarted == PR_TRUE) {
		    PR_Write(std_out, buf, nb);
		} else {
		    separateReqHeader(std_out, buf, nb, &wrStarted,
				      &headerSeparatorPtrnId);
		}
		if (verbose)
		    fputs("]]]\n\n", stderr);
		more = 1;
	    }
            break;
         case DEMO_DYING:
            FPRINTF(stderr, "%s: %d: closing fd %d\n", progName, ds->id, fd);
            close(fd);
            ds->fd = -1;
	    ds->state = DEMO_DEAD;
            break;
	}
    } while (more);

    if (verbose > 2) 
        fprintf(stderr, "%s: %d: crank_state_machine end\n", progName, ds->id);
}

static void socket_ready_callback(int socket, short flags, void* context)
{
    short mask;
    struct demo_session *ds = (struct demo_session *) context;

    FPRINTF(stderr, "%s: %d: socket_ready(%d, %x, )\n", progName, ds->id, 
            socket, flags);

    crank_state_machine(ds);

    mask = 0;
    if (memio_buffer_used(&ds->bufs->writebuf))
        mask |= EV_WRITE;
    if (memio_buffer_unused(&ds->bufs->readbuf))
        mask |= EV_READ;

    /* Re-add this socket to the list of watched sockets */
    event_set(&ds->e, socket, mask, socket_ready_callback, ds);
    event_add(&ds->e, NULL);
}

/*--------------------------------------------------------------------------*/

/* Native Unix networking functions. */
static int setNonBlocking(int fd)
{
    int flags = fcntl(fd, F_GETFL, 0);
    if (-1 == flags)
      flags = 0;
    return fcntl(fd, F_SETFL, flags | O_NONBLOCK);
}

/*--------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
    char *             host	=  NULL;
    char *             certDir  =  NULL;
    char *             tmp;
    int                multiplier = 0;
    SECStatus          rv;
    PRStatus           status;
    int                error = 0;
    PRUint16           portno = 443;
    PLOptState *optstate;
    PLOptStatus optstatus;
    int connections = 1;
    int numalive;
    struct demo_session demo_sessions[MAXSESSIONS];
    int i;

    std_out = PR_GetSpecialFD(PR_StandardOutput);

    progName = strrchr(argv[0], '/');
    if (!progName)
	progName = strrchr(argv[0], '\\');
    progName = progName ? progName+1 : argv[0];

    tmp = PR_GetEnv("NSS_DEBUG_TIMEOUT");
    if (tmp && tmp[0]) {
       int sec = PORT_Atoi(tmp);
       if (sec > 0) {
           maxInterval = PR_SecondsToInterval(sec);
       }
    }

    optstate = PL_CreateOptState(argc, argv, "23BTSfc:h:p:d:m:n:oqr:suvw:x");
    while ((optstatus = PL_GetNextOpt(optstate)) == PL_OPT_OK) {
	switch (optstate->option) {
	  case '?':
	  default : Usage(progName); 			break;

          case '2': disableSSL2 = 1; 			break;

          case '3': disableSSL3 = 1; 			break;

          case 'B': bypassPKCS11 = 1; 			break;

          case 'T': disableTLS  = 1; 			break;

          case 'S': skipProtoHeader = PR_TRUE;                 break;

          case 'C': cipherString = PORT_Strdup(optstate->value); break;

	  case 'c': connections = PORT_Atoi(optstate->value); break;

          case 'h': host = PORT_Strdup(optstate->value);	break;

          case 'f':  clientSpeaksFirst = PR_TRUE;       break;

          case 'd': certDir = PORT_Strdup(optstate->value);   break;

	  case 'm':
	    multiplier = atoi(optstate->value);
	    if (multiplier < 0)
	    	multiplier = 0;
	    break;

	  case 'n': nickname = PORT_Strdup(optstate->value);	break;

	  case 'o': override = 1; 			break;

	  case 'p': portno = (PRUint16)atoi(optstate->value);	break;

	  case 's': disableLocking = 1;                 break;

	  case 'u': enableSessionTickets = PR_TRUE;	break;

	  case 'v': verbose++;	 			break;

	  case 'r': renegotiate = atoi(optstate->value);	break;

	  case 'w':
		password = PORT_Strdup(optstate->value);
		useCommandLinePassword = PR_TRUE;
		break;

	  case 'x': useExportPolicy = 1; 		break;
	}
    }

    PL_DestroyOptState(optstate);

    if (optstatus == PL_OPT_BAD)
	Usage(progName);

    if (!host || !portno) 
    	Usage(progName);

    PR_Init( PR_SYSTEM_THREAD, PR_PRIORITY_NORMAL, 1);

    /* set our password function */
    if ( useCommandLinePassword ) {
	PK11_SetPasswordFunc(ownPasswd);
    } else {
    	PK11_SetPasswordFunc(SECU_GetModulePassword);
    }

    /* open the cert DB, the key DB, and the secmod DB. */
    if (!certDir) {
	certDir = SECU_DefaultSSLDir();	/* Look in $SSL_DIR */
	certDir = SECU_ConfigDirectory(certDir);
    } else {
	char *certDirTmp = certDir;
	certDir = SECU_ConfigDirectory(certDirTmp);
	PORT_Free(certDirTmp);
    }
    rv = NSS_Init(certDir);
    if (rv != SECSuccess) {
	SECU_PrintError(progName, "unable to open cert database");
#if 0
    rv = CERT_OpenVolatileCertDB(handle);
	CERT_SetDefaultCertDB(handle);
#else
	return 1;
#endif
    }
    handle = CERT_GetDefaultCertDB();

    /* set the policy bits true for all the cipher suites. */
    if (useExportPolicy)
	NSS_SetExportPolicy();
    else
	NSS_SetDomesticPolicy();

    /* all the SSL2 and SSL3 cipher suites are enabled by default. */
    if (cipherString) {
	/* disable all the ciphers, then enable the ones we want. */
	disableAllSSLCiphers();
    }

    status = PR_StringToNetAddr(host, &addr);
    if (status == PR_SUCCESS) {
    	addr.inet.port = PR_htons(portno);
    } else {
	/* Lookup host */
	PRAddrInfo *addrInfo;
	void       *enumPtr   = NULL;

	addrInfo = PR_GetAddrInfoByName(host, PR_AF_UNSPEC, 
	                                PR_AI_ADDRCONFIG | PR_AI_NOCANONNAME);
	if (!addrInfo) {
	    SECU_PrintError(progName, "error looking up host");
	    return 1;
	}
	do {
	    enumPtr = PR_EnumerateAddrInfo(enumPtr, addrInfo, portno, &addr);
	} while (enumPtr != NULL &&
		 addr.raw.family != PR_AF_INET &&
		 addr.raw.family != PR_AF_INET6);
	PR_FreeAddrInfo(addrInfo);
	if (enumPtr == NULL) {
	    SECU_PrintError(progName, "error looking up host address");
	    return 1;
	}
    }

    printHostNameAndAddr(host, &addr);

    event_init();

    /* Set up each of the simultaneous session: */
    for (i=0; i<connections; i++) {
        const int kBufSize = 16384;
        PRFileDesc *s;
        int fd;

	/* Create fake I/O socket that points at a pair of memio_buffers */
        s = memio_CreateIOLayer(kBufSize);
        if (s == NULL) {
	    SECU_PrintError(progName, "error creating socket");
	    return 1;
        }
        /* Grab pointer to buffers before that layer is buried */
        demo_sessions[i].bufs = (struct memio_secret *) s->secret;

	/* Create SSL state machine */
        /* Push SSL onto our fake I/O socket, set up SSL options */
        s = setup_client(s);
        SSL_SetURL(s, host);
        /* Tell SSL we're a client; needed if not letting NSPR do socket I/O */
        SSL_ResetHandshake(s, 0);

        /* Create native (not NSPR) socket */
        fd = socket(AF_INET, SOCK_STREAM, 0);
        setNonBlocking(fd);

	/* Set current interest mask and message pump for this event */
	event_set(&demo_sessions[i].e, fd, EV_READ|EV_WRITE, 
                  socket_ready_callback, &demo_sessions[i]);

	/* Add this socket to the list of monitored sockets. */
	event_add(&demo_sessions[i].e, NULL);

	/* Set up our application-level state machine and start it */
        demo_sessions[i].id = i;
        demo_sessions[i].fd = fd;
        demo_sessions[i].s = s;
        demo_sessions[i].state = DEMO_START;
        crank_state_machine(&demo_sessions[i]);
    }

    /* And let them all run together */
    do {
        /* Give CPU time to any session ready to do I/O */
        event_loop(EVLOOP_ONCE);

        /* Exit the loop if no session still reading */
        /* In a real program, we would compute this incrementally */
        numalive = 0;
        for (i=0; i<connections; i++) {
            if (demo_sessions[i].state != DEMO_DEAD)
		numalive++;
        }
        FPRINTF(stderr, "numalive %d\n", numalive);
    } while (numalive > 0);

    /* Clean them all up */
    for (i=0; i<connections; i++) {
        PR_Close(demo_sessions[i].s);
    }

    SSL_ClearSessionCache();
    if (NSS_Shutdown() != SECSuccess) {
        exit(1);
    }

    PR_Cleanup();
    return error;
}
